@@ -15,11 +15,12 @@ use constant DB_HEADSIZE_V1   => 124;
 use constant DB_SIG_1         => 0x9AA2D903;
 use constant DB_SIG_2_v1      => 0xB54BFB65;
 use constant DB_SIG_2_v2      => 0xB54BFB67;
-use constant DB_VER_DW        => 0x00030002;
+use constant DB_VER_DW_V1     => 0x00030002;
+use constant DB_VER_DW_V2     => 0x00030000; # recent KeePass is 0x0030001
 use constant DB_FLAG_RIJNDAEL => 2;
 use constant DB_FLAG_TWOFISH  => 8;
 
-our $VERSION = '2.02';
+our $VERSION = '2.03';
 my %locker;
 my $salsa20_iv = "\xe8\x30\x09\x4b\x97\x20\x5d\x2a";
 my $qr_date = qr/^(\d\d\d\d)-(\d\d)-(\d\d)[T ](\d\d):(\d\d):(\d\d)(\.\d+|)?Z?$/;
@@ -103,6 +104,7 @@ sub parse_db {
     $buffer = $$buffer if ref $buffer;
 
     my $head = $self->parse_header($buffer);
+    local $head->{'raw'} = substr $buffer, 0, $head->{'header_size'} if $head->{'version'} == 2;
     $buffer = substr $buffer, $head->{'header_size'};
 
     $self->unlock if $self->{'groups'}; # make sure we don't leave dangling keys should we reopen a new db
@@ -135,7 +137,7 @@ sub _parse_v1_header {
     my @f = qw(sig1 sig2 flags ver seed_rand enc_iv n_groups n_entries checksum seed_key rounds);
     my $t =   'L    L    L     L   a16       a16    L        L         a32      a32      L';
     @h{@f} = unpack $t, $buffer;
-    die "Unsupported file version ($h{'ver'}).\n" if $h{'ver'} & 0xFFFFFF00 != DB_VER_DW & 0xFFFFFF00;
+    die "Unsupported file version ($h{'ver'}).\n" if $h{'ver'} & 0xFFFFFF00 != DB_VER_DW_V1 & 0xFFFFFF00;
     $h{'enc_type'} = ($h{'flags'} & DB_FLAG_RIJNDAEL) ? 'rijndael'
                    : ($h{'flags'} & DB_FLAG_TWOFISH)  ? 'twofish'
                    : die "Unknown encryption type\n";
@@ -223,6 +225,16 @@ sub _parse_v2_body {
     $buffer = eval { $self->decompress($buffer) } or die "Failed to decompress document: $@" if ($head->{'compression'} || '') eq '1';
     $self->{'xml_in'} = $buffer if $self->{'keep_xml'} || $head->{'keep_xml'};
 
+    my $uuid = sub {
+        my $id = shift;
+        if ($id) {
+            $id = $self->decode_base64($id);
+            $id = 0 if $id eq "\0"x16;
+            $id =~ s/^0+(?=\d)// if $id =~ /^\d{16}$/;
+        }
+        return $id;
+    };
+
     # parse the XML - use our own parser since XML::Simple does not do event based actions
     my $tri = sub { return !defined($_[0]) ? undef : ('true' eq lc $_[0]) ? 1 : ('false' eq lc $_[0]) ? 0 : undef };
     my $s20_stream = $self->salsa20_stream({key => sha256($head->{'protected_stream_key'}), iv => $salsa20_iv, rounds => 20});
@@ -232,7 +244,7 @@ sub _parse_v2_body {
     my $level = 0;
     my $data = $self->parse_xml($buffer, {
         top            => 'KeePassFile',
-        force_array    => {map {$_ => 1} qw(Binaries Binary Group Entry String Association Item)},
+        force_array    => {map {$_ => 1} qw(Binaries Binary Group Entry String Association Item DeletedObject)},
         start_handlers => {Group => sub { $level++ }},
         end_handlers   => {
             Meta => sub {
@@ -247,6 +259,9 @@ sub _parse_v2_body {
                     $META->{lc $copy} = $copy =~ /_changed$/i ? $self->_parse_v2_date($node->{$key}) : $node->{$key};
                 }
                 $META->{'recycle_bin_enabled'} = $tri->($META->{'recycle_bin_enabled'});
+                $META->{$_} = $uuid->($META->{$_}) for qw(entry_templates_group last_selected_group last_top_visible_group recycle_bin_uuid);
+                die "HeaderHash recorded in file did not match actual hash of header.\n"
+                    if $META->{'header_hash'} && $head->{'raw'} && $META->{'header_hash'} ne $self->encode_base64(sha256($head->{'raw'}));
             },
             Binary => sub {
                 my ($node, $parent, $parent_tag, $tag) = @_;
@@ -273,7 +288,7 @@ sub _parse_v2_body {
             Group => sub {
                 my ($node, $parent, $parent_tag) = @_;
                 my $group = {
-                    id       => $node->{'UUID'},
+                    id       => $uuid->($node->{'UUID'}),
                     icon     => $node->{'IconID'},
                     title    => $node->{'Name'},
                     expanded => $tri->($node->{'IsExpanded'}),
@@ -286,7 +301,7 @@ sub _parse_v2_body {
                     auto_type_default => $node->{'DefaultAutoTypeSequence'},
                     auto_type_enabled => $tri->($node->{'EnableAutoType'}),
                     enable_searching  => $tri->($node->{'EnableSearching'}),
-                    last_top_entry    => $node->{'LastTopVisibleEntry'},
+                    last_top_entry    => $uuid->($node->{'LastTopVisibleEntry'}),
                     expires_enabled   => $tri->($node->{'Times'}->{'Expires'}),
                     location_changed  => $self->_parse_v2_date($node->{'Times'}->{'LocationChanged'}),
                     usage_count       => $node->{'Times'}->{'UsageCount'},
@@ -298,7 +313,6 @@ sub _parse_v2_body {
                 if ($parent_tag eq 'Group') {
                     push @{ $parent->{'__groups__'} }, $group;
                 } else {
-                    $group->{'__parent_tag__'} = $parent_tag;
                     push @GROUPS, $group;
                 }
             },
@@ -316,7 +330,7 @@ sub _parse_v2_body {
                     modified => $self->_parse_v2_date($node->{'Times'}->{'LastModificationTime'}),
                     comment  => delete($str{'Notes'}),
                     icon     => $node->{'IconID'},
-                    id       => $node->{'UUID'},
+                    id       => $uuid->($node->{'UUID'}),
                     title    => delete($str{'Title'}),
                     url      => delete($str{'URL'}),
                     username => delete($str{'UserName'}),
@@ -357,9 +371,24 @@ sub _parse_v2_body {
                 my ($node, $parent, $parent_tag, $tag) = @_;
                 $parent->{$tag} = delete($node->{'__entries__'}) || [];
             },
+            Association => sub {
+                my ($node, $parent) = @_;
+                push @{ $parent->{'__auto_type__'} },  {window => $node->{'Window'}, keys => $node->{'KeystrokeSequence'}};
+            },
+            DeletedObject => sub {
+                my ($node) = @_;
+                push @{ $GROUPS[0]->{'deleted_objects'} }, {
+                    uuid => $self->decode_base64($node->{'UUID'}),
+                    date => $self->_parse_v2_date($node->{'DeletionTime'}),
+                } if $GROUPS[0] && $node->{'UUID'} && $node->{'DeletionTime'};
+            },
         },
     });
 
+    my $g = $GROUPS[0];
+    @GROUPS = @{ $g->{'groups'} } if @GROUPS == 1
+        && $g && $g->{'notes'} && $g->{'notes'} eq "Added as a top group by File::KeePass"
+        && @{ $g->{'groups'} || [] } && !@{ $g->{'entries'} || [] } && !$g->{'auto_type_default'};
     return ($META, \@GROUPS);
 }
 
@@ -431,7 +460,7 @@ sub _parse_v1_entries {
         die "Entry header offset is out of range for type $type. ($pos, ".length($buffer).", $size)" if $pos + $size > length($buffer);
 
         if ($type == 1) {
-            $entry->{'id'}        = unpack 'H*', substr($buffer, $pos, $size);
+            $entry->{'id'}        = substr($buffer, $pos, $size);
         } elsif ($type == 2) {
             $entry->{'group_id'}  = unpack 'L', substr($buffer, $pos, 4);
         } elsif ($type == 3) {
@@ -466,7 +495,7 @@ sub _parse_v1_entries {
             if (!$ref) { # orphaned nodes go in special group
                 $gid = -1;
                 if (!$gmap->{$gid}) {
-                    push @$groups, ($gmap->{$gid} = {id => $gid, title => '*Orphaned*', icon => 0});
+                    push @$groups, ($gmap->{$gid} = {id => $gid, title => '*Orphaned*', icon => 0, created => $self->now});
                 }
                 $ref = $gmap->{$gid};
             }
@@ -630,14 +659,17 @@ sub _gen_v1_db {
     my $buffer  = '';
     my $entries = '';
     my %gid;
-    my $gid = sub { # v1 groups id size can only be a 32 bit int - not an arbitrary string like in v2
+    my $gid = sub { # v1 groups id size can only be a 32 bit int - v2 is supposed to be a 16 digit string
         local $_ = my $gid = shift;
         return $gid{$gid} ||= do {
-            $_ = unpack "N", substr /^[a-z0-9\/+]+=*$/i ? $self->decode_base64($_) : $_, 0, 4 if /\D/ || $_ > 2**32-1;
-            $_ = rand 2**32 while $gid{"\e$_\e"}++;
+            $_ = (/^\d+$/ && $_ < 2**32) ? $_ : /^([a-f0-9]{16})/i ? hex($1) : int(rand 2**32);
+            $_ = int(rand 2**32) while $gid{"\e$_\e"}++;
             $_;
         };
     };
+    my %uniq;
+    my $uuid = sub { return $self->uuid(shift, \%uniq) };
+
     my @g = $self->find_groups({}, $groups);
     if (grep {$_->{'expanded'}} @g) {
         my $bin = pack 'L', scalar(@g);
@@ -647,7 +679,7 @@ sub _gen_v1_db {
             title    => 'Meta-Info',
             username => 'SYSTEM',
             url      => '$',
-            id       => '00000000000000000000000000000000',
+            id       => '0000000000000000',
             group    => $g[0],
             binary   => {'bin-stream' => $bin},
         });
@@ -689,8 +721,7 @@ sub _gen_v1_db {
                 $txt .= "Auto-Type-Window".($i>1 ? "-$i" : '').": $_\n" for @{ $AT{$AT[$i-1]} };
             }
             my $com = defined($e->{'comment'}) ? "$txt$e->{'comment'}" : $txt;
-
-            my @d = ([1,      pack('LH*', length($e->{'id'})/2, $e->{'id'})],
+            my @d = ([1,      pack('L', 16). $uuid->($e->{'id'})],
                      [2,      pack('LL', 4, $gid->($g->{'id'}))],
                      [3,      pack('LL', 4, $e->{'icon'} || 0)],
                      [4,      pack('L', length($e->{'title'})+1)."$e->{'title'}\0"],
@@ -724,7 +755,7 @@ sub _gen_v1_header {
     $head->{'sig1'}  = DB_SIG_1;
     $head->{'sig2'}  = DB_SIG_2_v1;
     $head->{'flags'} = DB_FLAG_RIJNDAEL;
-    $head->{'ver'}   = DB_VER_DW;
+    $head->{'ver'}   = DB_VER_DW_V1;
     $head->{'n_groups'}  ||= 0;
     $head->{'n_entries'} ||= 0;
     die "Length of $_ was not 32 (".length($head->{$_}).")\n" for grep {length($head->{$_}) != 32} qw(seed_key checksum);
@@ -754,12 +785,18 @@ sub _gen_v2_db {
     if ($head->{'sig2'} && $head->{'sig2'} eq DB_SIG_2_v1) {
         $head->{'seed_rand'} = $head->{'seed_rand'}x2 if $head->{'seed_rand'} && length($head->{'seed_rand'}) == 16; # if coming from a v1 db augment the key (roundtripable)
     }
+    $head->{'compression'} = 1 if ! defined $head->{'compression'};
+    $head->{'start_bytes'} ||= join '', map {chr rand 256} 1 .. 32;
+    $head->{'protected_stream_key'} ||= join '', map {chr rand 256} 1..32;
     my $key = $self->_master_key($pass, $head);
+    my $header = $self->_gen_v2_header($head);
+
     my $buffer  = '';
+    my $untri = sub { return (!defined($_[0]) && !$_[1]) ? 'null' : !$_[0] ? 'False' : 'True' };
+    my %uniq;
+    my $uuid = sub { my $id = (defined($_[0]) && $_[0] eq '0') ? "\0"x16 : $self->uuid($_[0], \%uniq); return $self->encode_base64($id) };
 
-    my $untri = sub { return !defined($_[0]) ? 'null' : !$_[0] ? 'False' : 'True' };
-
-    my @mfld = qw(Generator DatabaseName DatabaseNameChanged DatabaseDescription DatabaseDescriptionChanged DefaultUserName DefaultUserNameChanged
+    my @mfld = qw(Generator HeaderHash DatabaseName DatabaseNameChanged DatabaseDescription DatabaseDescriptionChanged DefaultUserName DefaultUserNameChanged
                         MaintenanceHistoryDays Color MasterKeyChanged MasterKeyChangeRec MasterKeyChangeForce MemoryProtection
                         RecycleBinEnabled RecycleBinUUID RecycleBinChanged EntryTemplatesGroup EntryTemplatesGroupChanged HistoryMaxItems HistoryMaxSize
                         LastSelectedGroup LastTopVisibleGroup Binaries CustomData);
@@ -774,34 +811,33 @@ sub _gen_v2_db {
         $META->{$k} = $self->_gen_v2_date($META->{$k}) if $k =~ /Changed$/;
     };
     my $now = $self->_gen_v2_date;
+    $META->{'HeaderHash'} = $self->encode_base64(sha256($header));
     $def->(Color                      => '');
-    $def->(CustomData                 => {});
     $def->(DatabaseDescription        => '');
     $def->(DatabaseDescriptionChanged => $now, $qr_date);
     $def->(DatabaseName               => '');
     $def->(DatabaseNameChanged        => $now, $qr_date);
     $def->(DefaultUserName            => '');
     $def->(DefaultUserNameChanged     => $now, $qr_date);
-    $def->(EntryTemplatesGroup        => '');
     $def->(EntryTemplatesGroupChanged => $now, $qr_date);
     $def->(Generator                  => ref($self));
     $def->(HistoryMaxItems            => 10, qr{^\d+$});
     $def->(HistoryMaxSize             => 6291456, qr{^\d+$});
-    $def->(LastSelectedGroup          => '');
-    $def->(LastTopVisibleGroup        => '');
     $def->(MaintenanceHistoryDays     => 365, qr{^\d+$});
     $def->(MasterKeyChangeForce       => -1);
     $def->(MasterKeyChangeRec         => -1);
     $def->(MasterKeyChanged           => $now, $qr_date);
     $def->(RecycleBinChanged          => $now, $qr_date);
-    $def->(RecycleBinUUID             => '');
-    $META->{'RecycleBinEnabled'}      = $untri->(exists($META->{'RecycleBinEnabled'}) ? $META->{'RecycleBinEnabled'} : 1);
+    $META->{$_} = $uuid->($META->{$_} || 0) for qw(EntryTemplatesGroup LastSelectedGroup LastTopVisibleGroup RecycleBinUUID);
+    $META->{'RecycleBinEnabled'}      = $untri->(exists($META->{'RecycleBinEnabled'}) ? $META->{'RecycleBinEnabled'} : 1, 1);
     my $p = $META->{'MemoryProtection'} ||= {};
     for my $new (qw(ProtectTitle ProtectUserName ProtectPassword ProtectURL ProtectNotes)) { # unflatten protection
         (my $key = lc $new) =~ s/protect/protect_/;
         push @{$p->{'__sort__'}}, $new;
         $p->{$new} = (exists($META->{$key}) ? delete($META->{$key}) : ($key eq 'protect_password')) ? 'True' : 'False';
     }
+    my $cd = $META->{'CustomData'} ||= {};
+    $META->{'CustomData'} = {Item => [map {{Key => $_, Value => $cd->{$_}}} sort keys %$cd]} if ref($cd) eq 'HASH' && scalar keys %$cd;
 
     my @GROUPS;
     my $BIN = $META->{'Binaries'}->{'Binary'} = [];
@@ -820,14 +856,14 @@ sub _gen_v2_db {
         my ($e, $parent) = @_;
         push @$parent, my $E = {
             __sort__ => [qw(UUID IconID ForegroundColor BackgroundColor OverrideURL Tags Times String AutoType History)],
-            UUID   => $e->{'id'} || join('',map {sprintf '%02X', rand 256} 10),
+            UUID   => $uuid->($e->{'id'}),
             IconID => $e->{'icon'} || 0,
             Times  => {
                 __sort__             => [qw(LastModificationTime CreationTime LastAccessTime ExpiryTime Expires UsageCount LocationChanged)],
-                Expires              => $untri->($e->{'expires_enabled'}),
+                Expires              => $untri->($e->{'expires_enabled'}, 1),
                 UsageCount           => $e->{'usage_count'} || 0,
                 LastAccessTime       => $self->_gen_v2_date($e->{'accessed'}),
-                ExpiryTime           => $self->_gen_v2_date($e->{'expires'}),
+                ExpiryTime           => $self->_gen_v2_date($e->{'expires'} || $self->default_exp),
                 CreationTime         => $self->_gen_v2_date($e->{'created'}),
                 LastModificationTime => $self->_gen_v2_date($e->{'modified'}),
                 LocationChanged      => $self->_gen_v2_date($e->{'location_changed'}),
@@ -835,10 +871,10 @@ sub _gen_v2_db {
             Tags            => $e->{'tags'},
             BackgroundColor => $e->{'background_color'},
             ForegroundColor => $e->{'foreground_color'},
-            CustomIconUUID  => $e->{'custom_icon_uuid'},
+            CustomIconUUID  => $uuid->($e->{'custom_icon_uuid'} || 0),
             OverrideURL     => $e->{'override_url'},
             AutoType        => {
-                Enabled     => $untri->($e->{'auto_type_enabled'}),
+                Enabled     => $untri->(exists($e->{'auto_type_enabled'}) ? $e->{'auto_type_enabled'} : 1, 1),
                 DataTransferObfuscation => $e->{'auto_type_munge'} ? 1 : 0,
             },
         };
@@ -886,27 +922,27 @@ sub _gen_v2_db {
         return if ref($group) ne 'HASH';
         push @$parent, my $G = {
             __sort__ => [qw(UUID Name Notes IconID Times IsExpanded DefaultAutoTypeSequence EnableAutoType EnableSearching LastTopVisibleEntry)],
-            UUID   => $group->{'id'} || join('',map {sprintf '%02X', rand 256} 10),
+            UUID   => $uuid->($group->{'id'}),
             Name   => $group->{'title'} || '',
             Notes  => $group->{'notes'},
             IconID => $group->{'icon'} || 0,
             Times  => {
                 __sort__             => [qw(LastModificationTime CreationTime LastAccessTime ExpiryTime Expires UsageCount LocationChanged)],
-                Expires              => $untri->($group->{'expires_enabled'}),
+                Expires              => $untri->($group->{'expires_enabled'}, 1),
                 UsageCount           => $group->{'usage_count'} || 0,
                 LastAccessTime       => $self->_gen_v2_date($group->{'accessed'}),
-                ExpiryTime           => $self->_gen_v2_date($group->{'expires'}),
+                ExpiryTime           => $self->_gen_v2_date($group->{'expires'} || $self->default_exp),
                 CreationTime         => $self->_gen_v2_date($group->{'created'}),
                 LastModificationTime => $self->_gen_v2_date($group->{'modified'}),
                 LocationChanged      => $self->_gen_v2_date($group->{'location_changed'}),
             },
-            IsExpanded              => $untri->($group->{'expanded'}),
+            IsExpanded              => $untri->($group->{'expanded'}, 1),
             DefaultAutoTypeSequence => $group->{'auto_type_default'},
-            EnableAutoType          => lc($untri->($group->{'auto_type_enabled'})),
-            EnableSearching         => lc($untri->($group->{'enable_searching'})),
-            LastTopVisibleEntry     => $group->{'last_top_entry'},
+            EnableAutoType          => lc($untri->(exists($group->{'auto_type_enabled'}) ? $group->{'auto_type_enabled'} : 1)),
+            EnableSearching         => lc($untri->(exists($group->{'enable_searching'}) ? $group->{'enable_searching'} : 1)),
+            LastTopVisibleEntry     => $uuid->($group->{'last_top_entry'} || 0),
         };
-        $G->{'CustomIconUUID'} = $group->{'custom_icon_uuid'} if $group->{'custom_icon_uuid'}; # TODO
+        $G->{'CustomIconUUID'} = $uuid->($group->{'custom_icon_uuid'}) if $group->{'custom_icon_uuid'}; # TODO
         push @{$G->{'__sort__'}}, 'Entry' if @{ $group->{'entries'} || [] };
         foreach my $e (@{ $group->{'entries'} || [] }) {
             $gen_entry->($e, $G->{'Entry'} ||= []);
@@ -914,9 +950,18 @@ sub _gen_v2_db {
         push @{$G->{'__sort__'}}, 'Group' if @{ $group->{'groups'} || [] };
         $rec->($_, $G->{'Group'} ||= []) for @{ $group->{'groups'} || []};
     };
+    $groups = [{title => "Database", groups => [@$groups], notes => "Added as a top group by File::KeePass", expanded => 1}] if @$groups > 1;
     $rec->($_, \@GROUPS) for @$groups;
 
-    $head->{'protected_stream_key'} ||= join '', map {chr rand 256} 1..32;
+    if (@$groups && $groups->[0]->{'deleted_objects'}) {
+        foreach my $dob (@{ $groups->[0]->{'deleted_objects'} }) {
+            push @{ $data->{'Root'}->{'DeletedObjects'}->{'DeletedObject'} }, {
+                UUID => $self->encode_base64($dob->{'uuid'}),
+                DeletionTime => $self->_gen_v2_date($dob->{'date'}),
+            }
+        }
+    }
+
     my $s20_stream = $self->salsa20_stream({key => sha256($head->{'protected_stream_key'}), iv => $salsa20_iv, rounds => 20});
     for my $ref (@PROTECT_BIN, @PROTECT_STR) {
         $$ref = $self->encode_base64($s20_stream->($$ref));
@@ -930,19 +975,18 @@ sub _gen_v2_db {
         sort => {
             AutoType => [qw(Enabled DataTransferObfuscation Association)],
             Association => [qw(Window KeystrokeSequence)],
+            DeletedObject => [qw(UUID DeletionTime)],
         },
         no_trailing_newline => 1,
     });
     $self->{'xml_out'} = $buffer if $self->{'keep_xml'} || $head->{'keep_xml'};
 
-    $head->{'compression'} = 1 if ! defined $head->{'compression'};
     $buffer = $self->compress($buffer) if $head->{'compression'} eq '1';
     $buffer = $self->chunksum($buffer);
 
-    $head->{'start_bytes'} ||= join '', map {chr rand 256} 1 .. 32;
     substr $buffer, 0, 0, $head->{'start_bytes'};
 
-    return $self->_gen_v2_header($head) . $self->encrypt_rijndael_cbc($buffer, $key, $head->{'enc_iv'});
+    return $header . $self->encrypt_rijndael_cbc($buffer, $key, $head->{'enc_iv'});
 }
 
 sub _gen_v2_date {
@@ -956,9 +1000,9 @@ sub _gen_v2_header {
     my ($self, $head) = @_;
     $head->{'sig1'}        = DB_SIG_1;
     $head->{'sig2'}        = DB_SIG_2_v2;
-    $head->{'ver'}         = DB_VER_DW;
+    $head->{'ver'}         = DB_VER_DW_V2;
     $head->{'comment'}     = '' if ! defined $head->{'comment'};
-    $head->{'compression'} = (!$head->{'compression'} || $head->{'compression'} eq '1') ? 1 : 0;
+    $head->{'compression'} = (!defined($head->{'compression'}) || $head->{'compression'} eq '1') ? 1 : 0;
     $head->{'0'}           ||= "\r\n\r\n";
     $head->{'protected_stream_key'} ||= join '', map {chr rand 256} 1..32;
     die "Missing start_bytes\n" if ! $head->{'start_bytes'};
@@ -966,6 +1010,7 @@ sub _gen_v2_header {
     die "Length of enc_iv was not 16\n" if length($head->{'enc_iv'}) != 16;
 
     my $buffer = pack 'L3', @$head{qw(sig1 sig2 ver)};
+
     my $pack = sub { my ($type, $str) = @_; $buffer .= pack('C S', $type, length($str)) . $str };
     $pack->(1, $head->{'comment'}) if defined($head->{'comment'}) && length($head->{'comment'});
     $pack->(2, "\x31\xc1\xf2\xe6\xbf\x71\x43\x50\xbe\x58\x05\x21\x6a\xfc\x5a\xff"); # aes cipher
@@ -1022,7 +1067,7 @@ sub unchunksum {
         my ($index, $hash, $size) = unpack "\@$pos L a32 i", $buffer;
         $pos += 40;
         if ($size == 0) {
-            warn "Found mismatch for 0 chunksize\n" if $hash !~ /^\x00{32}$/;
+            warn "Found mismatch for 0 chunksize\n" if $hash ne "\0"x32;
             last;
         }
         #print "$index $hash $size\n";
@@ -1042,10 +1087,11 @@ sub chunksum {
     my $pos = 0;
     while ($pos < length($buffer)) {
         my $chunk = substr($buffer, $pos, $chunk_size);
-        $new .= pack("L a32 i", $index++, sha256($chunk), length($chunk));
+        $new .= pack "L a32 i", $index++, sha256($chunk), length($chunk);
         $new .= $chunk;
         $pos += length($chunk);
     }
+    $new .= pack "L a32 i", $index++, "\0"x32, 0;
     return $new;
 }
 
@@ -1203,6 +1249,21 @@ sub escape_xml {
     return $_;
 }
 
+sub uuid {
+    my ($self, $id, $uniq) = @_;
+    $id = $self->gen_uuid if !defined($id) || !length($id);
+    return $uniq->{$id} ||= do {
+        if (length($id) != 16) {
+            $id = substr($self->encode_base64($id), 0, 16) if $id !~ /^\d+$/ || $id > 2**32-1;
+            $id = sprintf '%016s', $id if $id ne '0';
+        }
+        $id = $self->gen_uuid while $uniq->{$id}++;
+        $id;
+    };
+}
+
+sub gen_uuid { shift->encode_base64(join '', map {chr rand 256} 1..12) } # (3072 bit vs 4096) only 8e28 entries vs 3e38 - but readable
+
 ###----------------------------------------------------------------###
 
 sub dump_groups {
@@ -1258,16 +1319,18 @@ sub find_groups {
     my ($self, $args, $groups, $level) = @_;
     my @tests = $self->finder_tests($args);
     my @groups;
-    my %used;
+    my %uniq;
     my $container = $groups || $self->groups;
     for my $g (@$container) {
         $g->{'level'} = $level || 0;
         $g->{'title'} = '' if ! defined $g->{'title'};
         $g->{'icon'}  ||= 0;
-        while (!defined($g->{'id'}) || $used{$g->{'id'}}++) {
-            warn "Found duplicate group_id - generating new one for \"$g->{'title'}\"" if defined($g->{'id'});
-            $g->{'id'} = int(rand 2**32-1);
+        if ($self->{'force_v2_gid'}) {
+            $g->{'id'} = $self->uuid($g->{'id'}, \%uniq);
+        } else {
+            $g->{'id'} = int(rand 2**32-1) while !defined($g->{'id'}) || $uniq{$g->{'id'}}++; # the non-v2 gid is compatible with both v1 and our v2 implementation
         }
+
         if (!@tests || !grep{!$_->($g)} @tests) {
             push @groups, $g;
             push @{ $self->{'__group_groups'} }, $container if $self->{'__group_groups'};
@@ -1309,15 +1372,18 @@ sub add_entry {
         $group = $self->find_group({id => $group}, $groups) || die "Could not find a matching group to add entry to.\n";
     }
 
+    my %uniq;
+    foreach my $g ($self->find_groups({}, $groups)) {
+        $uniq{$_->{'id'}}++ for @{ $g->{'entries'} || [] };
+    }
+    $args->{'id'} = $self->uuid($args->{'id'}, \%uniq);
     $args->{$_} = ''         for grep {!defined $args->{$_}} qw(title url username password comment);
-    $args->{$_} = 0          for grep {!defined $args->{$_}} qw(id icon);
+    $args->{$_} = 0          for grep {!defined $args->{$_}} qw(icon);
     $args->{$_} = $self->now for grep {!defined $args->{$_}} qw(created accessed modified);
     $args->{'expires'} ||= $self->default_exp;
     $self->_check_v1_binary($args);
     $self->_check_v1_auto_type($args);
-    while (!$args->{'id'} || $args->{'id'} !~ /^[a-f0-9]{32}$/ || $self->find_entry({id => $args->{'id'}}, $groups)) {
-        $args->{'id'} = join '', map {sprintf '%2x', rand 256} 1..16;
-    }
+
 
     push @{ $group->{'entries'} ||= [] }, $args;
     return $args;
@@ -1838,7 +1904,7 @@ Groups will look similar to the following:
     $g = [{
          expanded => 0,
          icon     => 0,
-         id       => 234234234,
+         id       => 234234234, # under v1 this is a 32 bit int, under v2 it is a 16 char id
          title    => 'Foo',
          level    => 0,
          entries => [{
@@ -1852,7 +1918,7 @@ Groups will look similar to the following:
              password => 'somepass', # will be hidden if the database is locked
              url      => "",
              username => "someuser",
-             id       => "0a55ac30af68149f62c072d7cc8bd5ee"
+             id       => "0a55ac30af68149f", # v1 is any hex char, v2 is any 16 char
          }],
          groups => [{
              expanded => 0,
@@ -1937,7 +2003,7 @@ The following fields can be passed to both v1 and v2 databases.
     password => 'somepass', # will be hidden if the database is locked
     url      => "http://",
     username => "someuser",
-    id       => "0a55ac30af68149f62c072d7cc8bd5ee", # randomly generated automatically
+    id       => "0a55ac30af68149f", # auto generated if needed, v1 is any hex char, v2 is any 16 char
     group    => $gid, # which group to add the entry to
 
 For compatibility with earlier versions of File::KeePass, it is
